import os
import threading
import json

class SimpleMemory:
    def __init__(self, file_name):
        self.file_name = file_name
        self._lock = threading.Lock()

    def get_memory(self):
        if os.path.exists(self.file_name):
            with open(self.file_name, 'r', encoding='utf-8') as f:
                return json.load(f)
        else:
            return {}

    def get_memory_value(self, key, default = None):
        with self._lock:
            memory = self.get_memory()
            if key in memory:
                return memory[key]
            return default

    def save_memory_value(self, key, value):
        with self._lock:
            memory = self.get_memory()
            memory[key] = value
            with open(self.file_name, 'w', encoding='utf-8') as f:
                json.dump(memory, f, ensure_ascii=False)



import logging
import re


logger = logging.getLogger(__name__)


def parseToolResponse(response, tool_names):
    regex_pattern = r"(" + "|".join(item + ":" for item in tool_names) + ")"
    tools = re.split(regex_pattern, response)
    tools = [tool.strip() for tool in tools if tool.strip()]
    tools_dict = {}
    for i in range(1, len(tools), 2):
        tool_name = tools[i-1].strip(':').strip()
        tool_argument = tools[i].strip()
        tools_dict[tool_name] = tool_argument
        logger.info(tool_name + ": " + tool_argument)
    return tools_dict


import requests
import sseclient
import json


class LlamaCppBackend:
    def __init__(self, url, max_predict=1024):
        self.url = url
        self.max_predict = max_predict

    def get_request_object(self, request_tokens, stream, temp, top_p):
        return {"prompt": request_tokens,
                "stream": stream,
                "n_predict": self.max_predict,
                "temperature": temp,
                "repeat_last_n": 0,
                "repeat_penalty": 1.0,
                "top_k": -1,
                "top_p": top_p,
                "min_p": 0,
                "tfs_z": 1,
                "typical_p": 1,
                "presence_penalty": 0,
                "frequency_penalty": 0,
                "stop": [self.stop_token, self.tokenizer.eos_token],
                "cache_prompt": True}

    def completion(self, request_tokens, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, False, temp, top_p)
        response = requests.post(self.url, json=request)
        response.raise_for_status()
        return response.json()["content"]

    async def stream_completion(self, request_tokens, callback, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, True, temp, top_p)
        response = requests.post(self.url, json=request, stream=True, headers={'Accept': 'text/event-stream'})
        response.raise_for_status()
        stream = sseclient.SSEClient(response).events()
        text_resp = ""
        for event in stream:
            parsed_event = json.loads(event.data)
            if parsed_event["stop"]:
                break
            content = parsed_event["content"]
            text_resp += content
            await callback(content)
        return text_resp


class TogetherAiBackend:
    def __init__(self, api_token, max_predict=1024):
        self.url = "https://api.together.xyz/v1/completions"
        self.max_predict = max_predict
        self.api_token = api_token

    def get_request_object(self, request_tokens, stream, temp, top_p):
        print(self.stop_token)
        return {"model": self.base_model,
                "prompt": self.tokenizer.decode(request_tokens),
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "max_tokens": self.max_predict,
                "stop": [self.stop_token, self.tokenizer.eos_token],
                "temperature": temp,
                "top_p": top_p,
                "top_k": -1,
                "repetition_penalty": 1,
                "stream": stream,
                "min_p": 0}

    def completion(self, request_tokens, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, False, temp, top_p)
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_token}"
        }
        response = requests.post(self.url, json=request, headers=headers, verify=False)
        response.raise_for_status()
        return response.json()["choices"][0]["text"]

    async def stream_completion(self, request_tokens, callback, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, True, temp, top_p)
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_token}"
        }
        response = requests.post(self.url, json=request, stream=True, headers=headers, verify=False)
        response.raise_for_status()
        stream = sseclient.SSEClient(response).events()
        text_resp = ""
        for event in stream:
            event_data = event.data
            parsed_event = json.loads(event_data)
            if parsed_event["choices"][0]["finish_reason"] is not None:
                break
            content = parsed_event["choices"][0]["text"]
            text_resp += content
            await callback(content)
        return text_resp


class DeepInfraBackend:
    def __init__(self, api_token, max_predict=1024):
        self.url = "https://api.deepinfra.com/v1/inference/"
        self.max_predict = max_predict
        self.api_token = api_token

    def get_request_object(self, request_tokens, stream, temp, top_p):
        if top_p == 0:
            top_p = 0.1
        return {"input": self.tokenizer.decode(request_tokens),
                "max_new_tokens": self.max_predict,
                "stop": [self.stop_token, self.tokenizer.eos_token],
                "temperature": temp,
                "top_p": top_p,
                "top_k": 0,
                "repetition_penalty": 1,
                "presence_penalty": 0,
                "frequency_penalty": 0,
                "stream": stream,
                "min_p": 0}

    def completion(self, request_tokens, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, False, temp, top_p)
        headers = {
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_token}"
        }
        response = requests.post(self.url + self.base_model, json=request, headers=headers, verify=False)
        response.raise_for_status()
        return response.json()["results"][0]["generated_text"]

    async def stream_completion(self, request_tokens, callback, temp=0.5, top_p=0.5):
        request = self.get_request_object(request_tokens, True, temp, top_p)
        headers = {
            "content-type": "application/json",
            "accept": "text/event-stream",
            "Authorization": f"Bearer {self.api_token}"
        }
        response = requests.post(self.url + self.base_model, json=request, stream=True, headers=headers, verify=False)
        response.raise_for_status()
        stream = sseclient.SSEClient(response).events()
        text_resp = ""
        for event in stream:
            event_data = event.data
            parsed_event = json.loads(event_data)
            if parsed_event["details"] != None and parsed_event["details"]["finish_reason"] != None:
                break
            content = parsed_event["token"]["text"]
            text_resp += content
            await callback(content)
        return text_resp



import locale
import pytz

from datetime import datetime


def get_time_string(tz = 'Europe/Moscow'):
    locale.setlocale(locale.LC_ALL, ('ru_RU', 'UTF-8'))
    dt = datetime.now(pytz.timezone(tz))
    return dt.strftime(" ### Текущее время: %A, %d %B, %Yг. %H:%M")



import os
import json
import logging
from transformers import AutoTokenizer, AutoConfig


logger = logging.getLogger(__name__)


class Aisha:
    def __init__(self, llm_backend, base_model, max_context=4096, prompt="", prompt_file=""):
        self.tokenizer = AutoTokenizer.from_pretrained(base_model, add_bos_token=False)
        self.max_context = max_context
        self.max_predict = llm_backend.max_predict

        if prompt_file:
            with open(prompt_file) as f:
                prompt = f.read()

        config = AutoConfig.from_pretrained(base_model)

        match config.model_type:
            case "qwen2":
                self.generation_promp_template = "<|im_start|>assistant\n"
                self.user_req_template = "<|im_start|>user\n{user_req}<|im_end|>"
                self.system_injection_template = "<|im_start|>system\n{system_injection}<|im_end|>"
                self.tokens = [self.tokenizer.apply_chat_template([{"role": "system", "content": prompt}])]
                self.stop_token = "<|im_end|>"
            case "llama":
                self.generation_promp_template = "<|start_header_id|>assistant<|end_header_id|>\n"
                self.user_req_template = "<|start_header_id|>user<|end_header_id|>\n{user_req}<|eot_id|>"
                self.system_injection_template = "<|start_header_id|>system<|end_header_id|>\n{system_injection}<|eot_id|>"
                self.tokens = [self.tokenizer.apply_chat_template([{"role": "system", "content": prompt}])]
                self.stop_token = "<|eot_id|>"
            case "cohere":
                self.generation_promp_template = "<|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"
                self.user_req_template = "<|START_OF_TURN_TOKEN|><|USER_TOKEN|>{user_req}<|END_OF_TURN_TOKEN|>"
                self.system_injection_template = "<|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>{system_injection}<|END_OF_TURN_TOKEN|>"
                self.tokens = [self.tokenizer.apply_chat_template([{"role": "system", "content": prompt}])]
                self.stop_token = self.tokenizer.eos_token
            case "phi3":
                self.generation_promp_template = "<|assistant|>\n"
                self.user_req_template = "<|user|>\n{user_req}<|end|>\n"
                self.system_injection_template = "<|system|>\n{system_injection}<|end|>\n"
                self.tokens = [self.tokenizer(self.tokenizer.bos_token + f"<|system|>\n{prompt}<|end|>\n")["input_ids"]]
                self.stop_token = "<|end|>"
            case "gemma2":
                self.generation_promp_template = "<start_of_turn>model\n"
                self.user_req_template = "<start_of_turn>user\n{user_req}<end_of_turn>\n"
                self.system_injection_template = "<start_of_turn>system\n{system_injection}<end_of_turn>\n"
                self.tokens = [self.tokenizer(self.tokenizer.bos_token + f"<start_of_turn>system\n{prompt}<end_of_turn>\n")["input_ids"]]
                self.stop_token = "<end_of_turn>"
            case _:
                raise RuntimeError("Unknown model: " + config.model_type)

        self.llm_backend = llm_backend
        self.llm_backend.stop_token = self.stop_token
        self.llm_backend.base_model = base_model
        self.llm_backend.tokenizer = self.tokenizer

        self.generation_prompt_tokens = self.tokenizer(self.generation_promp_template)["input_ids"]
        logger.info("System prompt size: " + str(len(self.tokens[0])))

    def sanitize(self, text):
        return text.replace("#", "").replace("<|", "").replace("|>", "")

    def add_user_request(self, user_request, system_injection="", meta_info=""):
        text = self.user_req_template.replace("{user_req}", self.sanitize(user_request.strip()) + meta_info)
        if system_injection:
            text += self.system_injection_template.replace("{system_injection}", system_injection)
        tokens = self.tokenizer(text)["input_ids"]
        self.tokens.append(tokens)
        self._cut_context()  # Освобождаем место под ответ модели

    def add_system_injection(self, system_injection):
        text = self.system_injection_template.replace("{system_injection}", system_injection)
        self.tokens.append(self.tokenizer(text)["input_ids"])
        self._cut_context()  # Освобождаем место под ответ модели

    def completion(self, temp=0.0, top_p=0.5):
        request_tokens = sum(self.tokens, [])
        request_tokens += self.generation_prompt_tokens
        text_resp = self.llm_backend.completion(request_tokens, temp, top_p)
        response_tokens = self.tokenizer(text_resp.strip() + self.stop_token)["input_ids"]
        response_tokens = self.generation_prompt_tokens + response_tokens
        self.tokens.append(response_tokens)
        return text_resp

    async def stream_completion(self, callback, temp=0.0, top_p=0.5):
        request_tokens = sum(self.tokens, [])
        request_tokens += self.generation_prompt_tokens
        text_resp = await self.llm_backend.stream_completion(request_tokens, callback, temp, top_p)
        response_tokens = self.tokenizer(text_resp.strip() + self.stop_token)["input_ids"]
        response_tokens = self.generation_prompt_tokens + response_tokens
        self.tokens.append(response_tokens)
        return text_resp

    def load_context(self, file_name):
        if os.path.isfile(file_name):
            with open(file_name) as f:
                self.tokens = json.load(f)

    def save_context(self, file_name):
        with open(file_name, "w") as f:
            json.dump(self.tokens, f)

    def _cut_context(self):
        busy_tokens = len(sum(self.tokens, []))
        free_tokens = self.max_context - busy_tokens
        if free_tokens < self.max_predict:
            while free_tokens < self.max_predict * 8:  # обрезаем с большим запасом, чтобы кеш контекста работал лучше
                free_tokens += len(self.tokens[1])
                del self.tokens[1]

    def clear_context(self):
        del self.tokens[1:]
